/* Modulo de controle da populacao */



#include <stdlib.h>
#include <stdio.h>

#include "defs.h"

#include "pop.h"
#include "reprod.h"
#include "aval.h"



/* Seleciona um individuo.
Baseado em "Roulette Wheel Parent Selection".
pfSomas - somas dos valores de avaliacao dos individuos com os individuos
          anteriores no vetor 
retorna indice do individuo selecionado
**********************************************************/
unsigned int SelecionaIndividuo ( float* pfSomas )
{
   // gera um numero entre 0 e soma total da avaliacao da populacao
   float
      fValorSelec = (float) ( rand () % (int) ( pfSomas [ TAM_POP - 1 ] * 10000.0 ) );
   fValorSelec /= 10000.0;

   // procura um valor maior ou igual ao sorteado, se encontrar seleciona este individuo
   for ( int ncP = 0; ncP < TAM_POP; ncP ++ )
      if ( pfSomas [ ncP ] >= fValorSelec )
         return ncP;

   // pelo menos deve sempre encontrar o ultimo, ou eh bug
   printf ( "BUG: %s %d", __FILE__, __LINE__ );
   exit ( 0 );
   return 0;
}



/* Cria uma geracao com novos cromossomos filho
pchPais - pais, retorna a nova geracao criada a partir dos pais
**********************************************************/
void CriaNovaGeracao ( char* pachPais )
{
   // populacao de filhos, criada temporariamente para fins de selecao
   char
      achFilhos [ TAM_POP * NUM_BITS ];

   // avalia os pais
   // soma da avaliacao de cada um dos pais com os pais anteriores no array
   // Ex.: [10, 3, 7]    Somas parciais: 10, 13, 20
   float
      afSomas [ TAM_POP ];   
   float
      fSomaParcial = 0;    // a soma com os anteriores
   // para cada individuo - define seu valor na avaliacao e a soma com os anteriores
   for ( unsigned int ucI = 0; ucI < TAM_POP; ucI ++ ) {

      fSomaParcial += Avalia ( pachPais + ucI * NUM_BITS );
      afSomas [ ucI ] = fSomaParcial;
   }  

   // gera os filhos de acordo com os pais
   for ( unsigned int ucG = 0; ucG < TAM_POP - 1; ucG += 2 ) {

      // define quais pais usar na reproducao
      char
         *pachPai1 = pachPais + SelecionaIndividuo ( afSomas ) * NUM_BITS,
         *pachPai2 = pachPais + SelecionaIndividuo ( afSomas ) * NUM_BITS;
         
      // copia pais para area dos novos filhos
      memcpy ( (void*) ( achFilhos +   ucG       * NUM_BITS ), (void*) pachPai1, NUM_BITS );
      memcpy ( (void*) ( achFilhos + ( ucG + 1 ) * NUM_BITS ), (void*) pachPai2, NUM_BITS );

      // altera informacoes dos filhos
      Reproduz ( achFilhos + ucG * NUM_BITS, achFilhos + ( ucG + 1 ) * NUM_BITS );
   }

   // copia os filhos para pais (substituindo a geracao anterior)
   memcpy ( (void*) pachPais, (void*) achFilhos, TAM_POP * NUM_BITS );
}



/* Inicia a populacao randomicamente
pachPopulacao - a populacao
**********************************************************/
void IniciaPopulacao ( char* pachPopulacao )
{
   for ( int ncI = 0; ncI < TAM_POP * NUM_BITS; ncI ++ )
      *( pachPopulacao ++ ) = rand () % 2;      // atribui valor 0 ou 1; incrementa ponteiro
}



/* Faz interacoes que evoluem a populacao
uTotalSubst - numero total de substituicoes de individuos que devem
              ocorrer ateh o algoritmo parar
pchPop - a populacao
**********************************************************/
void EvoluiPopulacao ( unsigned int uTotalSubst, char* pachPopulacao )
{
   unsigned int
      uNumSubst = 0;    // quantas substituicoes jah ocorreram

   // cria geracoes - faz interacao entre cromossomos
   do {

      // cria uma geracao com novos cromossomos filho
      CriaNovaGeracao ( pachPopulacao );

   	uNumSubst += TAM_POP;

      float
         fMelhor,
         fMedia;
      EstatisticaPopulacao ( pachPopulacao, fMelhor, fMedia );
   	printf ( "\nDurante %d - melhor individuo %f, media %f", uNumSubst, fMelhor, fMedia );

   } while ( uNumSubst < uTotalSubst ); // Enquanto nao ocorreram tantas
                                        // substituicoes quantas deveria.
}



/* Procura numa populacao o melhor individuo e calcula a media das avaliacoes.
pachPopulacao - a populacao onde procurar o melhor individuo
fAvalMelhor - retorna o valor do individuo que obteve o resultado
mais alto na avaliacao.
fMediaPop - retorna a media da populacao
**********************************************************/
void EstatisticaPopulacao ( char* pachPopulacao, float& fAvalMelhor, float& fMediaPop )
{
   fAvalMelhor = MUITO_PEQUENO;
   fMediaPop = 0;

   // varre populacao procurando algum maior
   for ( unsigned int ucP = 0; ucP < TAM_POP; ucP ++ ) {
      
      // verifica se eh maior que o valor atualmente selecionado
      float
         fAvalAtual = Avalia ( pachPopulacao + ucP * NUM_BITS );
      if ( fAvalAtual > fAvalMelhor )
         fAvalMelhor = fAvalAtual;
         
      fMediaPop += fAvalAtual;
   }
   
   fMediaPop /= TAM_POP;
}
