/* Modulo de avaliacao */



#include <math.h>
#include <stdio.h>

#include "defs.h"



/* Converte a sequencia de bits do cromossomo em 2 numeros
pchSeq - aponta sequencia de bits 
fX1 - primeiro numero
fY1 - segundo numero
**********************************************************/
void ConverteCromossomo ( char* pchSeq, double& fX1, double& fY1 )
{
   // varre sequencia de bits e converte
   unsigned int
      uNumero1 = 0,       // recebe primeira metade de bits
      uNumero2 = 0;       // recebe segunda metade de bits
   for ( int ncB = 0; ncB < NUM_BITS / 2; ncB ++ ) {
      uNumero1 += *( pchSeq + NUM_BITS / 2 - ncB - 1 ) << ncB;
      uNumero2 += *( pchSeq + NUM_BITS     - ncB - 1 ) << ncB;
   }
   
   // converte numeros para double
   fX1 = (double) uNumero1;
   fY1 = (double) uNumero2;
}



/* Faz avaliacao usando Binary f6
pchCromossomo - aponta sequencia de bits que forma o cromossomo a ser avaliado
**********************************************************/
float Avalia ( char* pchCromossomo )
{
   // converte sequencia de bits para valores numericos (valores estao entre 0 .. 2^22-1)
   double
      fX1,
      fY1;
   ConverteCromossomo ( pchCromossomo, fX1, fY1 );
   
   // divide valores (valores estao entre 0 .. 200)
   #define VAL 0.00004768372718899898
   double
      fX2 = fX1 * VAL,
      fY2 = fY1 * VAL;
      
   // subtrai (valores estao entre -100 .. +100)
   double
      fX3 = fX2 - 100,
      fY3 = fY2 - 100;
      
   // define o valor para a funcao F6
   double
      fSupQuad = sin ( sqrt ( fX3 * fX3 + fY3 * fY3 ) ),
      fInfQuad = fX3 * fX3 + fY3 * fY3,
      fValorF6 = /* 0.5 - */ ( ( fSupQuad * fSupQuad ) / ( 1.0 + 0.001 * fInfQuad * fInfQuad ) );

   return (float) fValorF6;
}

