/* Modulo de controle da populacao */



#include <stdlib.h>
#include <stdio.h>

#include "defs.h"

#include "pop.h"
#include "reprod.h"
#include "aval.h"
#include "misc.h"



/* Exibe dados referentes a um individuo da populacao
afAvals - valores da avaliacao dos cromossos
pchPais - configuracao dos cromossomos
**********************************************************/
void ExibePopulacao ( float* pafAvals, char* pachPais )
{
   for ( unsigned int ucI = 0; ucI < TAM_POP; ucI ++ ) {
      printf ( "\n%d)\nAvaliacao:  %f\n", ucI, * ( pafAvals + ucI ) );
      printf ( "Cromossomo: " );
      for ( int ncC = 0; ncC < NUM_BITS; ncC ++ )
         printf ( "%d", * ( pachPais + NUM_BITS * ucI + ncC ) );
   }
}




/* Procura numa populacao o pior individuo.
pachPopulacao - a populacao onde procurar
retorna um ponteiro para o pior individuo
**********************************************************/
char* PiorIndividuo ( char* pachPopulacao )
{
   char
      *pchIndividuo = NULL;
   float
      fAvalPior = - MUITO_PEQUENO;  // valor grande

   // varre populacao procurando algum menor
   for ( unsigned int ucP = 0; ucP < TAM_POP; ucP ++ ) {
      
      // verifica se eh menor que o valor atualmente selecionado
      float
         fAvalAtual = Avalia ( pachPopulacao + ucP * NUM_BITS );
      if ( fAvalAtual < fAvalPior ) {
         fAvalPior = fAvalAtual;
         pchIndividuo = pachPopulacao + ucP * NUM_BITS;
      }
   }
   
   return pchIndividuo;
}



/* Seleciona um individuo.
Baseado em "Roulette Wheel Parent Selection".
pfSomas - somas dos valores de avaliacao dos individuos com os individuos
          anteriores no vetor 
retorna indice do individuo selecionado
**********************************************************/
unsigned int SelecionaIndividuo ( float* pfSomas )
{
   // gera um numero entre 0 e soma total da avaliacao da populacao
   float
      fValorSelec = (float) ( rand () % (int) ( pfSomas [ TAM_POP - 1 ] ) );

   // procura um valor maior ou igual ao sorteado, se encontrar seleciona este individuo
   for ( int ncP = 0; ncP < TAM_POP; ncP ++ )
      if ( pfSomas [ ncP ] >= fValorSelec )
         return ncP;

   // pelo menos deve sempre encontrar o ultimo, ou eh bug
   printf ( "BUG: %s %d", __FILE__, __LINE__ );
   exit ( 0 );
   return 0;
}



/* Ordena os cromossomos conforme seus valores na avaliacao.
pafAvals - a avaliacao de cada cromossomo
pchCroms - os cromossomos a ordenar
Altera ambas tabelas.
**********************************************************/
void OrdenaPopulacao ( float* pafAvals, char* pachCroms )
{
   // ordena cada elemento
   for ( unsigned int ucI = 0; ucI < TAM_POP; ucI ++ ) {
 
      // ordena
      float
         afValMaior = MUITO_PEQUENO;
      unsigned int
         uNdxMaior = 0;

      // encontra o maior dentre os que ainda nao foram ordenados
      for ( unsigned int ucO = ucI; ucO < TAM_POP; ucO ++ )
         // se encontrou um elemento maior que o atual
         if ( afValMaior <= * ( pafAvals + ucO ) ) {
         
            afValMaior = * ( pafAvals + ucO );
            uNdxMaior = ucO;
         }
         
      // coloca o maior no lugar
      
      // troca valor da avaliacao
      float
         fTemp = * ( pafAvals + uNdxMaior );
      * ( pafAvals + uNdxMaior ) = * ( pafAvals + ucI );
      * ( pafAvals + ucI ) = fTemp;
      
      // troca o cromossomo
      for ( unsigned int ucB = 0; ucB < NUM_BITS; ucB ++ ) {

         // troca valor do bit
         char
            chBit = * ( pachCroms + uNdxMaior * NUM_BITS + ucB );
         * ( pachCroms + uNdxMaior * NUM_BITS + ucB ) = * ( pachCroms + ucI * NUM_BITS + ucB );
         * ( pachCroms + ucI * NUM_BITS + ucB ) = chBit;
      }
   }
}



/* Procura por um certo cromossomo numa populacao
pachIndiv - aponta o cromossomo a ser procurado
pachCroms - populacao onde procurar
retorna - ponteiro para o cromossomo na populacao (se encontrou)
          senao retorna NULL
**********************************************************/ 
char* ProcuraIndividuo ( char* pachIndiv, char* pachCroms )
{
   // para cada individuo da populacao
   for ( unsigned int ucC = 0; ucC < TAM_POP; ucC ++ ) {
   
      // flag que indica se o individuo eh igual (==1) ou nao (==0)
      char
         chIgual = 1;

      // para cada bit do cromosso
      for ( unsigned int ucB = 0; ucB < NUM_BITS; ucB ++ )

         // testa se bit do cromossomo eh diferente
         if ( pachIndiv [ ucB ] != pachCroms [ ucC * NUM_BITS + ucB ] )
            // se sim altera flag
            chIgual = 0;
            
      // se o cromossomo testado eh igual, retorna ponteiro para ele
      if ( chIgual )
         return pachCroms + ucC * NUM_BITS;
   }
   
   // populacao nao possui cromosso igual
   return NULL;
}
            


/* Cria uma geracao com novos cromossomos filho
pchPais - pais, retorna a nova geracao criada a partir dos pais
**********************************************************/
void CriaNovaGeracao ( char* pachPais )
{
   // populacao de filhos, criada temporariamente para fins de selecao
   char
      achFilhos [ TAM_POP * NUM_BITS ];

   // avalia os pais
   float
      // possui o valor da avaliacao de cada pai
      afAvals [ TAM_POP ];
   // para cada individuo - define seu valor na avaliacao
   for ( unsigned int ucI = 0; ucI < TAM_POP; ucI ++ )
      afAvals [ ucI ] = Avalia ( pachPais + ucI * NUM_BITS );
      
   // ordena os pais conforme seus valores na avaliacao
   OrdenaPopulacao ( afAvals, pachPais );
      
   // grau de adaptacao (fitness) de cada um dos pais
   float
      afSomas [ TAM_POP ];
   float
      fValAdaptacao = VAL_INI_ADAPTACAO,
      fSomaParcial = 0; 
   // para cada individuo - define seu valor de adaptacao (e soma com anterires)
   for ( unsigned int ucI = 0; ucI < TAM_POP; ucI ++ ) {
      fSomaParcial += fValAdaptacao;
      afSomas [ ucI ] = fSomaParcial;
      fValAdaptacao -= VAL_DESC_ADAPTACAO;
   }

   // copia os pais para filhos (para iniciar nova geracao)
   memcpy ( (void*) achFilhos, (void*) pachPais, TAM_POP * NUM_BITS );

   // gera os filhos de acordo com os pais
   unsigned int
      uNumFilhosUteis = 0;
   // enquanto nao forem criados tantos filhos quanto necessario
   while ( uNumFilhosUteis < NUM_SUBS_INDIV ) {

      // define quais pais usar na reproducao
      char
         *pachPai1 = pachPais + SelecionaIndividuo ( afSomas ) * NUM_BITS,
         *pachPai2 = pachPais + SelecionaIndividuo ( afSomas ) * NUM_BITS;
         
      // os novos cromossomos
      char
         achFilho1 [ NUM_BITS ],
         achFilho2 [ NUM_BITS ];
         
      // copia pais para area dos novos filhos
      memcpy ( (void*) achFilho1, (void*) pachPai1, NUM_BITS );
      memcpy ( (void*) achFilho2, (void*) pachPai2, NUM_BITS );

      // altera informacoes dos filhos
      Reproduz ( achFilho1, achFilho2 );
      
      // Coloca filhos no lugar dos piores cromossomos na tabela de filhos;
      // isso se nao houver cromossomo igual ao filho gerado.
      // tenta colocar filho 1
      if ( ProcuraIndividuo ( achFilho1, achFilhos ) == NULL ) {

         // coloca no lugar dos piores pais
         void
            *pvDest = (void*) achFilhos + ( TAM_POP - uNumFilhosUteis - 1 ) * NUM_BITS;
         memcpy ( (void*) pvDest, (void*) achFilho1, NUM_BITS );
         uNumFilhosUteis ++;
      }
      // tenta colocar filho 2 se jah nao foram colocados tantos filhos quanto necessario
      if ( uNumFilhosUteis < NUM_SUBS_INDIV &&
           ProcuraIndividuo ( achFilho2, achFilhos ) == NULL ) {

         // coloca no lugar dos piores pais
         void
            *pvDest = (void*) achFilhos + ( TAM_POP - uNumFilhosUteis - 1 ) * NUM_BITS;
         memcpy ( (void*) pvDest, (void*) achFilho2, NUM_BITS );
         uNumFilhosUteis ++;
      }
   }
   
   // copia os filhos para pais (substituindo a geracao anterior)
   memcpy ( (void*) pachPais, (void*) achFilhos, TAM_POP * NUM_BITS );
}



/* Inicia a populacao randomicamente
pchPop - a populacao
**********************************************************/
void IniciaPopulacao ( char* pachPopulacao )
{
   for ( int ncI = 0; ncI < TAM_POP * NUM_BITS; ncI ++ )
      *( pachPopulacao ++ ) = rand () % 2;      // atribui valor 0 ou 1; incrementa ponteiro
}



/* Faz interacoes que evoluem a populacao
uTotalSubst - numero total de substituicoes de individuos que devem
              ocorrer ateh o algoritmo parar
pachPopulacao - a populacao
**********************************************************/
void EvoluiPopulacao ( unsigned int uTotalSubst, char* pachPopulacao )
{
   unsigned int
      uNumSubst = 0;    // quantas substituicoes jah ocorreram

   // cria geracoes - faz interacao entre cromossomos
   do {

      // cria uma geracao com novos cromossomos filho
      CriaNovaGeracao ( pachPopulacao );

   	uNumSubst += NUM_SUBS_INDIV;

      float
         fMelhor,
         fMedia;
      EstatisticaPopulacao ( pachPopulacao, fMelhor, fMedia );
   	printf ( "\nDurante %d - melhor individuo %f, media %f", uNumSubst, fMelhor, fMedia );

   } while ( uNumSubst < uTotalSubst ); // Enquanto nao ocorreram tantas substituicoes
                                        // quantas deveria.
}



/* Procura numa populacao o melhor individuo e calcula a media das avaliacoes.
pachPopulacao - a populacao onde procurar o melhor individuo
fAvalMelhor - retorna o valor do individuo que obteve o resultado
mais alto na avaliacao.
fMediaPop - retorna a media da populacao
**********************************************************/
void EstatisticaPopulacao ( char* pachPopulacao, float& fAvalMelhor, float& fMediaPop )
{
   fAvalMelhor = MUITO_PEQUENO;
   fMediaPop = 0;

   // varre populacao procurando algum maior
   for ( unsigned int ucP = 0; ucP < TAM_POP; ucP ++ ) {
      
      // verifica se eh maior que o valor atualmente selecionado
      float
         fAvalAtual = Avalia ( pachPopulacao + ucP * NUM_BITS );
      if ( fAvalAtual > fAvalMelhor )
         fAvalMelhor = fAvalAtual;
         
      fMediaPop += fAvalAtual;
   }
   
   fMediaPop /= TAM_POP;
}

