/* Modulo de reproducao */



#include <stdlib.h>
#include <stdio.h>

#include "defs.h"
#include "misc.h"



/* Executa mutacao sobre o cromossomo.
Para cada bit sorteia um valor, se passou no teste de
probabilidade, SORTEIA um novo valor para o bit.
pchCursor - aponta para o cromossomo que sera mutado
**********************************************************/
void Mutacao ( char* pchCursor )
{
   // percorre sequencia de bits
   for ( int ncB = 0; ncB < NUM_BITS; ncB ++ ) {
   
      // testa se deve mutar o bit
      if ( ( (float) ( rand () % 1000 ) / 1000 ) <= PROB_MUTAC )
         // muta (valor 0 ou 1)
         *pchCursor = rand () % 2;

      // proximo bit
      pchCursor ++;
   }
}



/* Troca os valores apontados
 PV1 e PV2 - ponteiros para os char's a serem trocados
**********************************************************/
void TrocaValores ( char* pchV1, char* pchV2 )
{
   char
      chIntermed = *pchV1;
   *pchV1 = *pchV2;
   *pchV2 = chIntermed;
}

      

/* Executa cross-over ("one-point crossover")
pchCromo1 e pchCromo2 - apontam para os cromossomos pais,
                        serao substituidos pelos filhos
**********************************************************/
void Crossover1Ponto ( char* pchCromo1, char* pchCromo2 )
{
   // sorteia um numero, se for escolhido fazer cross-over, executa
   if ( ( (float) ( rand () % 1000 ) / 1000 ) <= PROB_CROSS ) {

      int
         nPontoCorte = rand () % NUM_BITS;
   
      char
         *pchCursor1 = pchCromo1,
         *pchCursor2 = pchCromo2;
    
      // varre cromossomo fazendo a troca dos bits ateh
      // atingir o ponto de termino
      for ( int ncP = 0; ncP < nPontoCorte; ncP ++ ) {

         TrocaValores ( pchCursor1, pchCursor2 );
         pchCursor1 ++;
         pchCursor2 ++;
      }  
   }
}



/* Aplica reproducao sobre os cromossomos gerando filhos.
pchCromo1 e pchCromo2 - apontam para os cromossomos pais,
                        serao substituidos pelos filhos
retorna quantos cromossomos foram usados na reproducao, 1 ou 2
(se soh usou 1 entao o filho estah no segundo parametro).
**********************************************************/
unsigned int Reproduz ( char* pchCromo1, char* pchCromo2 )
{
   // define qual operador executar
   // os valores de adaptacao de cada operador, somados com anteriores
   float
      afAdapatacaoOperadores [ NUM_OPERADORES ] = { VAL_ADAPT_1PCO,
                                                    VAL_ADAPT_1PCO + VAL_ADAPT_MUTA };
   unsigned int
      // seleciona valor pelo "Roulette Wheel"
      uNumOperador = SelecionaValor ( NUM_OPERADORES, afAdapatacaoOperadores ),
      // numero de cromossomos usados no processo
      uNumCromosUsados = 0;
      
   // executa o devido operador
   switch ( uNumOperador ) {
   
      case 0:      // faz "one point cross-over"
         Crossover1Ponto ( pchCromo1, pchCromo2 );
         uNumCromosUsados = 2;
         break;
         
      case 1:      // faz mutacao
         // faz mutacao nos cromossomos
         Mutacao ( pchCromo2 );
         uNumCromosUsados = 1;
         break;
   }
   
   return uNumCromosUsados;
}
