/* Modulo de reproducao.
Observacao: a funcao que seleciona o operador usa variaveis estaticas, cuidado
ao usar esta funcao numa segunda serie de evolucoes (com outra populacao) */



#include <stdlib.h>
#include <stdio.h>

#include "defs.h"
#include "misc.h"



/* Executa mutacao sobre o cromossomo.
Para cada bit sorteia um valor, se passou no teste de
probabilidade, SORTEIA um novo valor para o bit.
pchCursor - aponta para o cromossomo que sera mutado
**********************************************************/
void Mutacao ( char* pchCursor )
{
   // percorre sequencia de bits
   for ( int ncB = 0; ncB < NUM_BITS; ncB ++ ) {
   
      // testa se deve mutar o bit
      if ( ( (float) ( rand () % 1000 ) / 1000 ) <= PROB_MUTAC )
         // muta (valor 0 ou 1)
         *pchCursor = rand () % 2;

      // proximo bit
      pchCursor ++;
   }
}



/* Troca os valores apontados
 PV1 e PV2 - ponteiros para os char's a serem trocados
**********************************************************/
void TrocaValores ( char* pchV1, char* pchV2 )
{
   char
      chIntermed = *pchV1;
   *pchV1 = *pchV2;
   *pchV2 = chIntermed;
}



/* Executa cross-over ("uniform crossover")
pchCromo1 e pchCromo2 - apontam para os cromossomos pais,
                        serao substituidos pelos filhos
**********************************************************/
void CrossoverUniforme ( char* pchCromo1, char* pchCromo2 )
{
   // sorteia um numero, se for escolhido fazer cross-over, executa
   if ( ( (float) ( rand () % 1000 ) / 1000 ) <= PROB_CROSS ) {
         
      char
         *pchCursor1 = pchCromo1,
         *pchCursor2 = pchCromo2;
    
      // varre cromossomo fazendo a troca dos bits
      for ( int ncP = 0; ncP < NUM_BITS; ncP ++ ) {
      
         // troca os bits entre os cromossomos
         if ( ( rand () % 2 ) == 1 )
            TrocaValores ( pchCursor1, pchCursor2 );

         // proximos bits
         pchCursor1 ++;
         pchCursor2 ++;
      }  
   }
}



/* Aplica reproducao sobre os cromossomos gerando filhos.
pchCromo1 e pchCromo2 - apontam para os cromossomos pais,
                        serao substituidos pelos filhos
retorna quantos cromossomos foram gerados na reproducao, 1 ou 2
(se soh 1, entao o filho estah no segundo parametro,
o primeiro eh inalterado).
**********************************************************/
unsigned int Reproduz ( char* pchCromo1, char* pchCromo2 )
{
   // define qual operador executar
   
   static float
      // os valores de adaptacao de cada operador
      afAdapatacaoOperadores [ NUM_OPERADORES ] = { VAL_INI_ADAPT_1PCO,
                                                    VAL_INI_ADAPT_MUTA };
   // taxa de variacao do fitness de cada operador a cada interacao
   static float  // eh usado o primeiro como referencia
      afIncVals [ NUM_OPERADORES ] = { ( VAL_FIN_ADAPT_1PCO - VAL_INI_ADAPT_1PCO ) / NUM_SUBSTITUICOES,
                                       ( VAL_FIN_ADAPT_MUTA - VAL_INI_ADAPT_MUTA ) / NUM_SUBSTITUICOES };
      
   // define adaptacoes somadas
   // adaptacao somados com anteriores
   float
      afSomas [ NUM_OPERADORES ];
   float
      fSomaParcial = 0; 
   // para cada elemento - define seu valor de adaptacao (e soma com anterires)
   for ( unsigned int ucI = 0; ucI < NUM_OPERADORES; ucI ++ ) {
      fSomaParcial += afAdapatacaoOperadores [ ucI ];
      afSomas [ ucI ] = fSomaParcial;
   }

   unsigned int
      // seleciona valor pelo "Roulette Wheel"
      uNumOperador = SelecionaValor ( NUM_OPERADORES, afSomas ),
      // numero de cromossomos usados e gerados no processo
      uNumCromosGerados = 0;
      
   // executa o operador
   
   switch ( uNumOperador ) {
   
      case 0:      // faz cross-over
         CrossoverUniforme ( pchCromo1, pchCromo2 );
         uNumCromosGerados = 2;
         break;
         
      case 1:      // faz mutacao
         // faz mutacao nos cromossomos
         Mutacao ( pchCromo2 );
         uNumCromosGerados = 1;
         break;
   }
   
   // altera pesos da adaptacao
   
   // faz interpolacao
   for ( unsigned int ucP = 0; ucP < NUM_OPERADORES; ucP ++ )
      afAdapatacaoOperadores [ ucP ] += afIncVals [ ucP ] * uNumCromosGerados;
   
   // retorna numero de cromossomos gerados

   return uNumCromosGerados;
}

