/* Modulo de reproducao.
Observacao: a funcao que seleciona o operador usa variaveis estaticas, cuidado
ao usar esta funcao numa segunda serie de evolucoes (com outra populacao) */



#include <stdlib.h>
#include <stdio.h>

#include "defs.h"
#include "misc.h"



/* Troca os valores apontados
 PV1 e PV2 - ponteiros para os char's a serem trocados
**********************************************************/
void TrocaValores ( char* pchV1, char* pchV2 )
{
   char
      chIntermed = *pchV1;
   *pchV1 = *pchV2;
   *pchV2 = chIntermed;
}



/* Executa crossover ("uniform crossover")
pchCromo1 e pchCromo2 - apontam para os cromossomos pais,
                        serao substituidos pelos filhos
**********************************************************/
void CrossoverUniforme ( char* pchCromo1, char* pchCromo2 )
{
   // sorteia um numero, se for escolhido fazer crossover, executa
   if ( ( (float) ( rand () % 1000 ) / 1000 ) <= PROB_CROSS_UNIF ) {
         
      char
         *pchCursor1 = pchCromo1,
         *pchCursor2 = pchCromo2;
    
      // varre cromossomo fazendo a troca dos bits
      for ( int ncP = 0; ncP < NUM_BITS; ncP ++ ) {
      
         // troca os bits entre os cromossomos
         if ( ( rand () % 2 ) == 1 )
            TrocaValores ( pchCursor1, pchCursor2 );

         // proximos bits
         pchCursor1 ++;
         pchCursor2 ++;
      }  
   }
}



/* Executa crossover por media ("average crossover")
pchCromo1 e pchCromo2 - apontam para os cromossomos pais.
retorna no segundo parametro o filho.
**********************************************************/
void CrossoverMedia ( char* pchCromo1, char* pchCromo2 )
{
   // sorteia um numero, se for escolhido fazer crossover, executa
   if ( ( (float) ( rand () % 1000 ) / 1000 ) <= PROB_CROSS_MEDIA ) {
   
      // converte para numerico os cromossomos
      unsigned int
         // primeiro pai
         uX1,
         uY1,
         // segundo pai
         uX2,
         uY2;
      ConvBinVal ( pchCromo1, uX1, uY1 );   
      ConvBinVal ( pchCromo2, uX2, uY2 );   
      
      // faz a media
      unsigned int
         uXM = ( uX1 + uX2 ) / 2,
         uYM = ( uY1 + uY2 ) / 2;
         
      // converte para sequencia binaria, alterando
      ConvValBin ( uXM, uYM, pchCromo2 );         
   }
}



/* Executa mutacao sobre o cromossomo: sorteia um novo valor
para um ou outro dos numeros que compoem o cromossomo, se
passou no teste de probabilidade.
pchCromo - aponta para o cromossomo que sera mutado
**********************************************************/
void Mutacao ( char* pchCromo )
{
   // testa se deve mutar
   if ( ( (float) ( rand () % 1000 ) / 1000 ) <= PROB_MUTAC ) {

      // converte para numerico
      unsigned int
         uX,
         uY;
      ConvBinVal ( pchCromo, uX, uY );   

      // escolhe um novo valor
      unsigned int
         uValNovo = rand () % MAIOR_PARA_SORTEIO;

      // substitui um dos campos, convertendo para sequencia binaria
      if ( rand () % 2 )    // define qual dos campos substituir
         ConvValBin ( uX, uValNovo, pchCromo );
      else
         ConvValBin ( uValNovo, uY, pchCromo );
   }
}



/* Soma um certo valor ao cromossomo.
uMax - faixa de valores que pode ser sorteada
fProb - probabilidade de executar operacao
pchCromo - cromossomo a ser alterado
O parametro uMax define o limite maximo para a escolha de um
numero aleatorio, zero eh o minimo.
O valor somado eh o numero aleatorio menor a sua metade
(isso para poder haver substracao, tambem).
**********************************************************/
void Gradiente ( unsigned int uMax, float fProb, char* pchCromo )
{
   // testa se deve executar operacao
   if ( ( (float) ( rand () % 1000 ) / 1000 ) <= fProb ) {
   
      // sorteia valor
      unsigned int
         uVal = rand () % uMax;

      // define valor para somar no campo do cromossomo
      int
         nSoma = uVal - uVal / 2;
         
      // converte cromossomo para numerico
      unsigned int
         uX,
         uY;
      ConvBinVal ( pchCromo, uX, uY );   

      // soma num dos campos e converte para sequencia binaria
      if ( rand () % 2 )    // define qual dos campos somar
         ConvValBin ( uX + nSoma, uY, pchCromo );
      else
         ConvValBin ( uX, uY + nSoma, pchCromo );
   }
}      

         

/* Aplica reproducao sobre os cromossomos gerando filhos.
pchCromo1 e pchCromo2 - apontam para os cromossomos pais,
                        serao substituidos pelos filhos
retorna quantos cromossomos foram gerados na reproducao, 1 ou 2
(se soh 1, entao o filho estah no segundo parametro,
o primeiro eh inalterado).
**********************************************************/
unsigned int Reproduz ( char* pchCromo1, char* pchCromo2 )
{
   // define qual operador executar
   
   static float
      // os valores de adaptacao de cada operador
      afAdapatacaoOperadores [ NUM_OPERADORES ] = { VAL_INI_ADAPT_UNIF,
                                                    VAL_INI_ADAPT_MEDIA,
                                                    VAL_INI_ADAPT_MUTA,
                                                    VAL_INI_ADAPT_GRAD_1,
                                                    VAL_INI_ADAPT_GRAD_2 };
   // taxa de variacao do fitness de cada operador a cada interacao
   static float  // eh usado o primeiro como referencia
      afIncVals [ NUM_OPERADORES ] = { ( VAL_FIN_ADAPT_UNIF - VAL_INI_ADAPT_UNIF ) / NUM_SUBSTITUICOES,
                                       ( VAL_FIN_ADAPT_MEDIA - VAL_INI_ADAPT_MEDIA ) / NUM_SUBSTITUICOES,
                                       ( VAL_FIN_ADAPT_MUTA - VAL_INI_ADAPT_MUTA ) / NUM_SUBSTITUICOES,
                                       ( VAL_FIN_ADAPT_GRAD_1 - VAL_INI_ADAPT_GRAD_1 ) / NUM_SUBSTITUICOES,
                                       ( VAL_FIN_ADAPT_GRAD_2 - VAL_INI_ADAPT_GRAD_2 ) / NUM_SUBSTITUICOES };

   // define adaptacoes somadas
   // adaptacao somados com anteriores
   float
      afSomas [ NUM_OPERADORES ];
   float
      fSomaParcial = 0; 
   // para cada elemento - define seu valor de adaptacao (e soma com anterires)
   for ( unsigned int ucI = 0; ucI < NUM_OPERADORES; ucI ++ ) {
      fSomaParcial += afAdapatacaoOperadores [ ucI ];
      afSomas [ ucI ] = fSomaParcial;
   }

   unsigned int
      // seleciona valor pelo "Roulette Wheel"
      uNumOperador = SelecionaValor ( NUM_OPERADORES, afSomas ),
      // numero de cromossomos usados e gerados no processo
      uNumCromosGerados = 0;
      
   // executa o operador
   
   switch ( uNumOperador ) {
   
      case 0:      // faz crossover uniforme
         CrossoverUniforme ( pchCromo1, pchCromo2 );
         uNumCromosGerados = 2;
         break;
   
      case 1:      // faz crossover pela media
         CrossoverMedia ( pchCromo1, pchCromo2 );
         uNumCromosGerados = 1;
         break;
         
      case 2:      // faz mutacao
         Mutacao ( pchCromo2 );
         uNumCromosGerados = 1;
         break;
         
      case 3:      // faz gradiente 1
         Gradiente ( FAIXA_VALS_GRAD_1, PROB_GRAD_1, pchCromo2 );
         uNumCromosGerados = 1;
         break;
         
      case 4:      // faz gradiente 2
         Gradiente ( FAIXA_VALS_GRAD_2, PROB_GRAD_2, pchCromo2 );
         uNumCromosGerados = 1;
         break;
   }
   
   // altera pesos da adaptacao
   
   // faz interpolacao
   for ( unsigned int ucP = 0; ucP < NUM_OPERADORES; ucP ++ )
      afAdapatacaoOperadores [ ucP ] += afIncVals [ ucP ] * uNumCromosGerados;
   
   // retorna numero de cromossomos gerados

   return uNumCromosGerados;
}

